@@ -11,6 +11,8 @@
 #include <vector>
 
 #include "big_int.h"
+#include "exceptions.h"
+#include "fraction.h"
 #include "utils.h"
 
 namespace py = pybind11;
@@ -110,9 +112,8 @@ static std::string to_repr(const Type& value) {
   return {stream.str()};
 }
 
-using BaseInt =
-    BigInt<std::conditional_t<sizeof(void*) == 8, std::uint32_t, std::uint16_t>,
-           '_'>;
+using BaseInt = cppbuiltins::BigInt<
+    std::conditional_t<sizeof(void*) == 8, std::uint32_t, std::uint16_t>, '_'>;
 
 static int int_to_sign(const py::int_& value) {
   PyLongObject* ptr = (PyLongObject*)value.ptr();
@@ -124,8 +125,8 @@ static std::vector<BaseInt::Digit> int_to_digits(const py::int_& value) {
   PyLongObject* ptr = (PyLongObject*)value.ptr();
   Py_ssize_t signed_size = Py_SIZE(ptr);
   std::size_t size = Py_ABS(signed_size) + (signed_size == 0);
-  return binary_digits_to_binary_base<digit, BaseInt::Digit, PyLong_SHIFT,
-                                      BaseInt::BINARY_SHIFT>(
+  return cppbuiltins::binary_digits_to_binary_base<
+      digit, BaseInt::Digit, PyLong_SHIFT, BaseInt::BINARY_SHIFT>(
       std::vector<digit>(ptr->ob_digit, ptr->ob_digit + size));
 }
 
@@ -138,13 +139,23 @@ static const char* pystr_to_ascii_c_str(const py::str& string) {
   return result;
 }
 
+static py::int_ object_to_py_long(const py::object& value) {
+  PyObject* result_ptr = PyNumber_Index(value.ptr());
+  if (!result_ptr) {
+    throw py::error_already_set();
+  }
+  return py::reinterpret_steal<py::int_>(result_ptr);
+}
+
 class Int : public BaseInt {
  public:
   Int() : BaseInt() {}
 
   explicit Int(const BaseInt& value) : BaseInt(value) {}
 
-  explicit Int(py::int_ value)
+  explicit Int(const py::object& value) : Int(object_to_py_long(value)) {}
+
+  explicit Int(const py::int_& value)
       : BaseInt(int_to_sign(value), int_to_digits(value)) {}
 
   Int(const py::str& value, std::size_t base)
@@ -155,12 +166,7 @@ class Int : public BaseInt {
   }
 
   Int operator%(const Int& divisor) const {
-    try {
-      return Int(BaseInt::operator%(divisor));
-    } catch (const std::range_error& exception) {
-      PyErr_SetString(PyExc_ZeroDivisionError, exception.what());
-      throw py::error_already_set();
-    }
+    return Int(BaseInt::operator%(divisor));
   }
 
   const Int& operator+() const { return *this; }
@@ -181,58 +187,36 @@ class Int : public BaseInt {
     return Int(BaseInt::operator-(other));
   }
 
-  double operator/(const BigInt& divisor) const {
-    try {
-      return BaseInt::operator/(divisor);
-    } catch (const std::range_error& exception) {
-      PyErr_SetString(PyExc_ZeroDivisionError, exception.what());
-      throw py::error_already_set();
-    }
+  Int operator/(const Int& divisor) const {
+    return Int(BaseInt::operator/(divisor));
   }
 
   Int abs() const { return Int(BaseInt::abs()); }
 
   PyLongObject* as_PyLong() const {
-    int sign = this->sign();
     const std::vector<BaseInt::Digit>& digits = this->digits();
-    std::vector<digit> result_digits = binary_digits_to_binary_base<
-        BaseInt::Digit, digit, BaseInt::BINARY_SHIFT, PyLong_SHIFT>(digits);
+    std::vector<digit> result_digits =
+        cppbuiltins::binary_digits_to_binary_base<
+            BaseInt::Digit, digit, BaseInt::BINARY_SHIFT, PyLong_SHIFT>(digits);
     PyLongObject* result = _PyLong_New(result_digits.size());
     std::memcpy(result->ob_digit, result_digits.data(),
                 sizeof(digit) * result_digits.size());
-    Py_SET_SIZE(result, Py_SIZE(result) * sign);
+    Py_SET_SIZE(result, is_positive()
+                            ? result_digits.size()
+                            : (is_negative()
+                                   ? -static_cast<ssize_t>(result_digits.size())
+                                   : 0));
     return result;
   }
 
-  py::tuple divmod(const Int& divisor) const {
-    Int quotient, remainder;
-    try {
-      BaseInt::divmod(divisor, quotient, remainder);
-    } catch (const std::range_error& exception) {
-      PyErr_SetString(PyExc_ZeroDivisionError, exception.what());
-      throw py::error_already_set();
-    }
-    return py::make_tuple(quotient, remainder);
-  }
-
-  Int floor_divide(const Int& divisor) const {
-    try {
-      return Int(BaseInt::floor_divide(divisor));
-    } catch (const std::range_error& exception) {
-      PyErr_SetString(PyExc_ZeroDivisionError, exception.what());
-      throw py::error_already_set();
-    }
-  }
-
   Int gcd(const Int& other) const { return Int(BaseInt::gcd(other)); }
 
   Py_hash_t hash() const {
-    int sign = this->sign();
     const std::vector<BaseInt::Digit>& digits = this->digits();
     if (digits.size() == 1) {
-      if (sign > 0)
+      if (is_positive())
         return digits[0];
-      else if (sign < 0)
+      else if (is_negative())
         return -static_cast<Py_hash_t>(digits[0] + (digits[0] == 1));
       else
         return 0;
@@ -245,230 +229,88 @@ class Int : public BaseInt {
       result += *position;
       if (result >= _PyHASH_MODULUS) result -= _PyHASH_MODULUS;
     }
-    result = result * sign;
+    result = is_negative() ? -result : result;
     result -= (result == std::numeric_limits<Py_uhash_t>::max());
     return static_cast<Py_hash_t>(result);
   }
 
   Int invmod(const Int& divisor) const { return Int(BaseInt::invmod(divisor)); }
 
-  bool is_one() const { return BaseInt::is_one(); }
+  Int power(const Int& exponent) const { return Int(BaseInt::power(exponent)); }
 
-  py::object pow(const Int& exponent,
-                 const Int* maybe_modulus = nullptr) const {
-    if (maybe_modulus == nullptr && exponent.sign() < 0) {
-      PyObject* result = PyFloat_Type.tp_as_number->nb_power(
-          (PyObject*)as_PyLong(), (PyObject*)exponent.as_PyLong(), Py_None);
-      if (!result) throw py::error_already_set();
-      return py::reinterpret_steal<py::object>(result);
-    }
-    return py::cast(Int(BaseInt::pow(exponent, maybe_modulus)));
+  Int power_modulo(const Int& exponent, const Int& modulus) const {
+    return Int(BaseInt::power_modulo(exponent, modulus));
   }
-
-  int sign() const { return BaseInt::sign(); }
 };
 
 static std::ostream& operator<<(std::ostream& stream, const Int& value) {
   return stream << C_STR(MODULE_NAME) "." INT_NAME "('" << value.repr() << "')";
 }
 
-static py::object pow(const py::float_& base, const py::float_& exponent) {
+namespace cppbuiltins {
+template <>
+class Gcd<Int> {
+ public:
+  Int operator()(ConstParameterFrom<Int> first,
+                 ConstParameterFrom<Int> second) const {
+    return first.gcd(second);
+  }
+};
+
+template <>
+double divide_as_double<Int>(ConstParameterFrom<Int> dividend,
+                             ConstParameterFrom<Int> divisor) {
+  return dividend.divide_as_double(divisor);
+}
+
+template <>
+bool is_negative<Int>(ConstParameterFrom<Int> value) {
+  return value.is_negative();
+}
+
+template <>
+bool is_one<Int>(ConstParameterFrom<Int> value) {
+  return value.is_one();
+}
+
+template <>
+bool is_positive<Int>(ConstParameterFrom<Int> value) {
+  return value.is_positive();
+}
+
+template <>
+Int power<Int>(ConstParameterFrom<Int> base, ConstParameterFrom<Int> exponent) {
+  return base.power(exponent);
+}
+}  // namespace cppbuiltins
+
+using Fraction = cppbuiltins::Fraction<Int>;
+
+static py::object power(const py::float_& base, const py::float_& exponent) {
   PyObject* result = PyNumber_Power(base.ptr(), exponent.ptr(), Py_None);
   if (!result) throw py::error_already_set();
   return py::reinterpret_steal<py::object>(result);
 }
 
-class Fraction {
- public:
-  Fraction() : _numerator(Int()), _denominator(Int(1)) {}
+static py::object power(const Int& base, const Fraction& exponent) {
+  return exponent.denominator().is_one()
+             ? py::cast(Fraction(base).power(exponent.numerator()))
+             : power(py::float_(double{base}), py::float_(double{exponent}));
+}
 
-  Fraction(const Int& numerator, const Int& denominator = Int(1))
-      : Fraction(numerator, denominator, std::true_type{}) {}
-
-  Fraction operator+(const Fraction& other) const {
-    return Fraction(
-        _numerator * other._denominator + _denominator * other._numerator,
-        _denominator * other._denominator);
+Py_hash_t hash_fraction(const Fraction& value) {
+  static const Int HASH_MODULUS{_PyHASH_MODULUS};
+  Py_hash_t result;
+  try {
+    result = (Int(value.numerator().abs().hash()) *
+              value.denominator().invmod(HASH_MODULUS))
+                 .hash();
+  } catch (const std::invalid_argument&) {
+    result = _PyHASH_INF;
   }
-
-  Fraction operator+(const Int& other) const {
-    return Fraction(_numerator + _denominator * other, _denominator);
-  }
-
-  operator bool() const { return bool(_numerator); }
-
-  operator double() const { return _numerator / _denominator; }
-
-  bool operator==(const Fraction& other) const {
-    return _numerator == other._numerator && _denominator == other._denominator;
-  }
-
-  bool operator<=(const Fraction& other) const {
-    return sign() < other.sign() ||
-           (sign() == other.sign() &&
-            _numerator * other._denominator <= _denominator * other._numerator);
-  }
-
-  bool operator<(const Fraction& other) const {
-    return sign() < other.sign() ||
-           (sign() == other.sign() &&
-            _numerator * other._denominator < _denominator * other._numerator);
-  }
-
-  Fraction operator%(const Fraction& other) const {
-    return Fraction(
-        (_numerator * other._denominator) % (other._numerator * _denominator),
-        _denominator * other._denominator);
-  }
-
-  Fraction operator%(const Int& other) const {
-    return Fraction(_numerator % (other * _denominator), _denominator);
-  }
-
-  Fraction operator*(const Fraction& other) const {
-    const Int numerator_other_denominator_gcd =
-        _numerator.gcd(other._denominator);
-    const Int other_numerator_denominator_gcd =
-        _denominator.gcd(other._numerator);
-    return Fraction(
-        _numerator.floor_divide(numerator_other_denominator_gcd) *
-            other._numerator.floor_divide(other_numerator_denominator_gcd),
-        _denominator.floor_divide(other_numerator_denominator_gcd) *
-            other._denominator.floor_divide(numerator_other_denominator_gcd));
-  }
-
-  Fraction operator*(const Int& other) const {
-    const Int denominator_other_gcd = _denominator.gcd(other);
-    return Fraction(_numerator * other.floor_divide(denominator_other_gcd),
-                    _denominator.floor_divide(denominator_other_gcd));
-  }
-
-  Fraction operator-() const {
-    return Fraction(-_numerator, _denominator, std::false_type{});
-  }
-
-  const Fraction& operator+() const { return *this; }
-
-  Fraction operator-(const Fraction& other) const {
-    return Fraction(
-        _numerator * other._denominator - _denominator * other._numerator,
-        _denominator * other._denominator);
-  }
-
-  Fraction operator-(const Int& other) const {
-    return Fraction(_numerator - _denominator * other, _denominator);
-  }
-
-  Fraction operator/(const Fraction& other) const {
-    const Int numerators_gcd = _numerator.gcd(other._numerator);
-    const Int denominators_gcd = _denominator.gcd(other._denominator);
-    return Fraction(_numerator.floor_divide(numerators_gcd) *
-                        other._denominator.floor_divide(denominators_gcd),
-                    other._numerator.floor_divide(numerators_gcd) *
-                        _denominator.floor_divide(denominators_gcd));
-  }
-
-  Fraction operator/(const Int& other) const {
-    const Int numerators_gcd = _numerator.gcd(other);
-    return Fraction(_numerator.floor_divide(numerators_gcd),
-                    other.floor_divide(numerators_gcd) * _denominator);
-  }
-
-  Int ceil() const { return -(-_numerator).floor_divide(_denominator); }
-
-  Int floor() const { return _numerator.floor_divide(_denominator); }
-
-  const Int& denominator() const { return _denominator; }
-
-  py::tuple divmod(const Fraction& divisor) const {
-    Int quotient = floor_divide(divisor);
-    Fraction remainder = operator%(divisor);
-    return py::make_tuple(quotient, remainder);
-  }
-
-  Int floor_divide(const Fraction& other) const {
-    return (_numerator * other._denominator)
-        .floor_divide(other._numerator * _denominator);
-  }
-
-  Int floor_divide(const Int& other) const {
-    return _numerator.floor_divide(other * _denominator);
-  }
-
-  Py_hash_t hash() const {
-    static const Int HASH_MODULUS{_PyHASH_MODULUS};
-    Py_hash_t result;
-    try {
-      result =
-          (Int(_numerator.abs().hash()) * _denominator.invmod(HASH_MODULUS))
-              .hash();
-    } catch (const std::invalid_argument&) {
-      result = _PyHASH_INF;
-    }
-    if (_numerator.sign() < 0) result = -result;
-    return result - (result == -1);
-  }
-
-  const Int& numerator() const { return _numerator; }
-
-  Fraction pow(const Int& exponent) const {
-    if (exponent.sign() < 0) {
-      if (_numerator.sign() == 0) {
-        PyErr_SetString(PyExc_ZeroDivisionError,
-                        "Denominator should not be zero.");
-        throw py::error_already_set();
-      }
-      Int exponent_modulus = -exponent;
-      return _numerator.sign() < 0
-                 ? Fraction(Int(static_cast<BaseInt>(-_denominator)
-                                    .pow(exponent_modulus)),
-                            Int(static_cast<BaseInt>(-_numerator)
-                                    .pow(exponent_modulus)),
-                            std::false_type{})
-                 : Fraction(Int(static_cast<BaseInt>(_denominator)
-                                    .pow(exponent_modulus)),
-                            Int(static_cast<BaseInt>(_numerator)
-                                    .pow(exponent_modulus)),
-                            std::false_type{});
-    }
-    return Fraction(Int(static_cast<BaseInt>(_numerator).pow(exponent)),
-                    Int(static_cast<BaseInt>(_denominator).pow(exponent)),
-                    std::false_type{});
-  }
-
-  py::object pow(const Fraction& exponent) const {
-    return exponent._denominator.is_one()
-               ? py::cast(pow(exponent._numerator))
-               : ::pow(py::float_{double{*this}}, py::float_{double{exponent}});
-  }
-
-  int sign() const { return _numerator.sign(); }
-
-  operator Int() const { return sign() < 0 ? ceil() : floor(); }
-
- private:
-  Int _numerator, _denominator;
-
-  template <bool NORMALIZE>
-  Fraction(const Int& numerator, const Int& denominator,
-           std::bool_constant<NORMALIZE>)
-      : _numerator(numerator), _denominator(denominator) {
-    if constexpr (NORMALIZE) {
-      if (_denominator.sign() == 0) {
-        PyErr_SetString(PyExc_ZeroDivisionError,
-                        "Denominator should not be zero.");
-        throw py::error_already_set();
-      }
-      if (_denominator.sign() < 0) {
-        _numerator = -_numerator;
-        _denominator = -_denominator;
-      }
-      Int gcd = _numerator.gcd(_denominator);
-      _denominator = _denominator.floor_divide(gcd);
-      _numerator = _numerator.floor_divide(gcd);
-    }
-  }
-};
+  if (value.numerator().is_negative()) result = -result;
+  return result - (result == -1);
+}
 
 static Fraction operator+(const Int& self, const Fraction& other) {
   return Fraction(self * other.denominator() + other.numerator(),
@@ -487,26 +329,18 @@ static Fraction operator%(const Int& self, const Fraction& other) {
 
 static Fraction operator*(const Int& self, const Fraction& other) {
   const Int self_other_denominator_gcd = self.gcd(other.denominator());
-  return Fraction(
-      self.floor_divide(self_other_denominator_gcd) * other.numerator(),
-      other.denominator().floor_divide(self_other_denominator_gcd));
+  return Fraction((self / self_other_denominator_gcd) * other.numerator(),
+                  other.denominator() / self_other_denominator_gcd);
 }
 
 static Fraction operator/(const Int& self, const Fraction& other) {
   const Int self_other_numerator_gcd = self.gcd(other.numerator());
-  return Fraction(
-      self.floor_divide(self_other_numerator_gcd) * other.denominator(),
-      other.numerator().floor_divide(self_other_numerator_gcd));
+  return Fraction((self / self_other_numerator_gcd) * other.denominator(),
+                  other.numerator() / self_other_numerator_gcd);
 }
 
 Int floor_divide(const Int& self, const Fraction& other) {
-  return (self * other.denominator()).floor_divide(other.numerator());
-}
-
-py::object pow(const Int& base, const Fraction& exponent) {
-  return exponent.denominator().is_one()
-             ? py::cast(Fraction(base).pow(exponent.numerator()))
-             : pow(py::float_(double{base}), py::float_(double{exponent}));
+  return (self * other.denominator()) / other.numerator();
 }
 
 static std::ostream& operator<<(std::ostream& stream, const Fraction& value) {
@@ -549,7 +383,7 @@ class ListIterator {
     auto iterable_state = state[1].cast<IterableState>();
     auto raw = std::make_shared<RawList>();
     raw->reserve(iterable_state.size());
-    for (auto& element : iterable_state)
+    for (const auto&& element : iterable_state)
       raw->push_back(py::reinterpret_borrow<Object>(element));
     return {state[0].cast<Size>(), std::move(raw), state[2].cast<bool>()};
   }
@@ -583,7 +417,7 @@ class ListReversedIterator {
     auto iterable_state = state[1].cast<IterableState>();
     auto raw = std::make_shared<RawList>();
     raw->reserve(iterable_state.size());
-    for (auto& element : iterable_state)
+    for (const auto&& element : iterable_state)
       raw->push_back(py::reinterpret_borrow<Object>(element));
     return {state[0].cast<Size>(), std::move(raw), state[2].cast<bool>()};
   }
@@ -630,7 +464,7 @@ class List {
   static List from_state(IterableState state) {
     RawList raw;
     raw.reserve(state.size());
-    for (auto& element : state)
+    for (const auto&& element : state)
       raw.push_back(py::reinterpret_borrow<Object>(element));
     return {raw};
   }
@@ -1012,7 +846,7 @@ class Set {
   static Set from_state(IterableState state) {
     RawSet raw;
     raw.reserve(state.size());
-    for (auto& element : state)
+    for (const auto&& element : state)
       raw.insert(py::reinterpret_borrow<Object>(element));
     return {raw};
   }
@@ -1120,7 +954,7 @@ class Set {
 
   Set difference(py::args others) const {
     RawSet raw{*_raw};
-    for (const auto& other : others) {
+    for (const auto&& other : others) {
       RawSet values;
       fill_from_iterable(values, other.cast<py::iterable>());
       raw_sets_in_place_difference(raw, values);
@@ -1131,7 +965,7 @@ class Set {
   void difference_update(py::args others) {
     auto& raw = *_raw;
     auto size_before = raw.size();
-    for (const auto& other : others) {
+    for (const auto&& other : others) {
       RawSet values;
       fill_from_iterable(values, other.cast<py::iterable>());
       raw_sets_in_place_difference(raw, values);
@@ -1145,7 +979,7 @@ class Set {
 
   Set intersection(py::args others) const {
     RawSet raw{*_raw};
-    for (const auto& other : others) {
+    for (const auto&& other : others) {
       RawSet values;
       fill_from_iterable(values, other.cast<py::iterable>());
       raw_sets_in_place_intersection(raw, values);
@@ -1156,7 +990,7 @@ class Set {
   void intersection_update(py::args others) {
     auto& raw = *_raw;
     auto size_before = raw.size();
-    for (const auto& other : others) {
+    for (const auto&& other : others) {
       RawSet values;
       fill_from_iterable(values, other.cast<py::iterable>());
       raw_sets_in_place_intersection(raw, values);
@@ -1223,7 +1057,7 @@ class Set {
 
   Set union_(py::args others) const {
     RawSet raw{*_raw};
-    for (const auto& other : others) {
+    for (const auto&& other : others) {
       RawList values;
       fill_from_iterable(values, other.cast<py::iterable>());
       raw.insert(values.begin(), values.end());
@@ -1234,7 +1068,7 @@ class Set {
   void update(py::args others) {
     auto& raw = *_raw;
     auto size_before = raw.size();
-    for (const auto& other : others) {
+    for (const auto&& other : others) {
       RawList values;
       fill_from_iterable(values, other.cast<py::iterable>());
       raw.insert(values.begin(), values.end());
@@ -1269,13 +1103,23 @@ PYBIND11_MODULE(MODULE_NAME, m) {
       R"pbdoc(Alternative implementation of python builtins based on C++ `std` library.)pbdoc";
   m.attr("__version__") = C_STR(VERSION_INFO);
 
+  py::register_exception_translator([](std::exception_ptr ptr) {
+    try {
+      if (ptr) std::rethrow_exception(ptr);
+    } catch (const cppbuiltins::ZeroDivisionError& error) {
+      PyErr_SetString(PyExc_ZeroDivisionError, error.what());
+    }
+  });
+
   m.def("gcd", &Int::gcd);
 
+  static const Int ONE{1};
+
   py::class_<Int> PyInt(m, INT_NAME);
   PyInt.def(py::init<>())
       .def(py::init<const py::str&, std::size_t>(), py::arg("string"),
            py::arg("base") = 10)
-      .def(py::init<const py::int_&>(), py::arg("value"))
+      .def(py::init<const py::object&>(), py::arg("value"))
       .def(py::self + py::self)
       .def(py::self == py::self)
       .def(~py::self)
@@ -1286,36 +1130,65 @@ PYBIND11_MODULE(MODULE_NAME, m) {
       .def(-py::self)
       .def(+py::self)
       .def(py::self - py::self)
-      .def(py::self / py::self)
       .def(py::pickle([](const Int& self) { return py::int_(self); },
                       [](const py::int_& state) { return Int(state); }))
       .def("__abs__", &Int::abs)
       .def("__bool__", &Int::operator bool)
       .def("__ceil__", &identity<const Int&>)
       .def("__copy__", [](const Int& self) -> const Int& { return self; })
-      .def("__divmod__", &Int::divmod, py::is_operator{})
+      .def(
+          "__divmod__",
+          [](const Int& dividend, const Int& divisor) {
+            Int quotient, remainder;
+            dividend.divmod(divisor, quotient, remainder);
+            return py::make_tuple(quotient, remainder);
+          },
+          py::is_operator{})
       .def("__deepcopy__",
            [](const Int& self, const py::dict&) -> Int { return self; })
       .def("__float__", &Int::operator double)
       .def("__floor__", &identity<const Int&>)
-      .def("__floordiv__", &Int::floor_divide, py::is_operator{})
+      .def("__floordiv__", &Int::operator/, py::is_operator{})
       .def("__hash__", &Int::hash)
       .def("__int__", &Int::operator py::int_)
-      .def("__pow__", &Int::pow, py::arg("exponent"),
-           py::arg("modulus") = nullptr, py::is_operator{})
+      .def(
+          "__pow__",
+          [](const Int& base, const Int& exponent, const Int* maybe_modulus) {
+            if (maybe_modulus == nullptr && exponent.is_negative()) {
+              PyObject* result = PyFloat_Type.tp_as_number->nb_power(
+                  (PyObject*)base.as_PyLong(), (PyObject*)exponent.as_PyLong(),
+                  Py_None);
+              if (!result) throw py::error_already_set();
+              return py::reinterpret_steal<py::object>(result);
+            }
+            return py::cast(maybe_modulus == nullptr
+                                ? base.power(exponent)
+                                : base.power_modulo(exponent, *maybe_modulus));
+          },
+          py::arg("exponent"), py::arg("modulus") = nullptr, py::is_operator{})
       .def("__repr__", &to_repr<Int>)
       .def("__str__", &Int::repr<10>)
-      .def("__trunc__", &identity<const Int&>);
+      .def("__truediv__", &cppbuiltins::divide_as_double<Int>,
+           py::is_operator{})
+      .def("__trunc__", &identity<const Int&>)
+      .def_property_readonly("numerator", &identity<const Int&>)
+      .def_property_readonly("denominator",
+                             [](const Int&) -> const Int& { return ONE; });
 
   py::class_<Fraction> PyFraction(m, FRACTION_NAME);
   PyFraction.def(py::init<>())
       .def(py::init<const Int&, const Int&>(), py::arg("numerator"),
-           py::arg("denominator") = Int(1))
+           py::arg("denominator") = ONE)
       .def(py::self + py::self)
       .def(py::self + Int{})
       .def(py::self == py::self)
+      .def(py::self == Int{})
       .def(py::self <= py::self)
+      .def(py::self <= Int{})
+      .def(Int{} <= py::self)
       .def(py::self < py::self)
+      .def(py::self < Int{})
+      .def(Int{} < py::self)
       .def(py::self % py::self)
       .def(py::self % Int{})
       .def(py::self * py::self)
@@ -1341,7 +1214,15 @@ PYBIND11_MODULE(MODULE_NAME, m) {
           }))
       .def("__bool__", &Fraction::operator bool)
       .def("__ceil__", &Fraction::ceil)
-      .def("__divmod__", &Fraction::divmod, py::is_operator{})
+      .def(
+          "__divmod__",
+          [](const Fraction& dividend, const Fraction& divisor) {
+            Int quotient;
+            Fraction remainder;
+            dividend.divmod(divisor, quotient, remainder);
+            return py::make_tuple(quotient, remainder);
+          },
+          py::is_operator{})
       .def("__float__", &Fraction::operator double)
       .def("__floor__", &Fraction::floor)
       .def("__floordiv__",
@@ -1351,13 +1232,23 @@ PYBIND11_MODULE(MODULE_NAME, m) {
       .def("__floordiv__",
            py::overload_cast<const Int&>(&Fraction::floor_divide, py::const_),
            py::is_operator{})
-      .def("__hash__", &Fraction::hash)
+      .def("__hash__", &hash_fraction)
       .def("__int__", [](const Fraction& self) { return py::int_(Int(self)); })
-      .def("__pow__", py::overload_cast<const Int&>(&Fraction::pow, py::const_),
-           py::arg("exponent"), py::is_operator{})
-      .def("__pow__",
-           py::overload_cast<const Fraction&>(&Fraction::pow, py::const_),
-           py::arg("exponent"), py::is_operator{})
+      .def(
+          "__pow__",
+          [](const Fraction& base, const Int& exponent) {
+            return base.power(exponent);
+          },
+          py::arg("exponent"), py::is_operator{})
+      .def(
+          "__pow__",
+          [](const Fraction& base, const Fraction& exponent) {
+            return exponent.denominator().is_one()
+                       ? py::cast(base.power(exponent.numerator()))
+                       : power(py::float_{double{base}},
+                               py::float_{double{exponent}});
+          },
+          py::arg("exponent"), py::is_operator{})
       .def("__repr__", &to_repr<Fraction>)
       .def(
           "__rfloordiv__",
@@ -1368,7 +1259,7 @@ PYBIND11_MODULE(MODULE_NAME, m) {
       .def(
           "__rpow__",
           [](const Fraction& exponent, const Int& base) {
-            return pow(base, exponent);
+            return power(base, exponent);
           },
           py::is_operator{})
       .def("__str__",
